local libname = "MissionMinder";


local function newCallbackRegistry()
	local registry = {};
	local callbacks = LibStub("CallbackHandler-1.0"):New(registry)

	return {
		Register = function(self, module, event, func)
			if ( type(event) == "string" ) then
				registry.RegisterCallback(module, event, func)
			else
				for i = 1, #event do
					registry.RegisterCallback(module, event[i], func)
				end
			end
		end,

		Fire = function(self, event, ...)
			callbacks:Fire(event, ...);
		end
	}
end

local function newGlobal(variable)
	return {
		Set = function(self, table)
			_G[variable] = table
		end,

		Get = function(self)
			return _G[variable]
		end
	}
end

local DebugClass = {
	IsDebug = function(self)
		return false
	end;

	DEBUG = function(self, ...)
		if ( self:IsDebug() ) then
			print(...)
		end
	end;

	DEBUGF = function(self, ...)
		self:DEBUG(string.format(...))
	end;
}
DebugClass.__index = DebugClass

---------------------------------------------
-- Module
---------------------------------------------
local ModuleClass = {
	IsDebug = function(self)
		return self.lib.debug or self.addon.debug or self.debug
	end;

	RegisterEvent = function(self, event, func)
		self.lib.eventFrame:RegisterEvent(event)
		self.lib.callbacks.api:Register(self.fullname, event, func)
	end;

	WatchOption = function(self, name, func)
		self.lib.callbacks.options:Register(self.fullname, name, func)

		if ( self.lib.initialized ) then -- If already initialized, then immediately fire callback
			local options = self.lib.savedVariable:Get()
			self.lib.callbacks.options:Fire(name, options[name])
		end
	end;

	On = function(self, event, func)
		self.lib.callbacks.events:Register(self.fullname, event, func)
	end;

	Fire = function(self, event, ...)
		self:DEBUG("[INTERNAL EVENT]", event, ...)
		self.lib.callbacks.events:Fire(event, ...)
	end;
}
setmetatable(ModuleClass, DebugClass)
ModuleClass.__index = ModuleClass


---------------------------------------------
-- Addon
---------------------------------------------
local AddonClass = {
	IsDebug = function(self)
		return self.lib.debug or self.debug
	end;

	SetDefaultOptions = function(self, defaults)
		self.defaultOptions = defaults
	end;

	CreateModule = function(self, name)
		if ( self.modules[name] ) then
			return self.modules[name]
		end

		local module = {
			name = name;
			fullname = self.name .. "/" .. name;
			addon = self;
			lib = self.lib;

			initialize = function() end;
			initialized = false;

			debug = false;
		}
		setmetatable(module, ModuleClass)
		self.modules[name] = module

		return module
	end;
}
setmetatable(AddonClass, DebugClass)
AddonClass.__index = AddonClass


local function splitstr(str, delimiter)
	local elements = {}
	for i in str:gmatch("[^" .. delimiter .. "]+") do
		table.insert(elements, i)
	end

	return unpack(elements)
end

---------------------------------------------
-- Library
---------------------------------------------
local LibraryClass = {
	IsDebug = function(self)
		return self.debug
	end;

	SetDefaultOptions = function(self, defaults)
		self.defaultOptions = defaults
	end;

	GetOption = function(self, name)
		return self.savedVariable:Get()[name]
	end;

	SetOption = function(self, name, value)
		self.savedVariable:Get()[name] = value
		self:DEBUG("[OPTION CHANGED]", name, value)
		self.callbacks.options:Fire(name, value)
	end;

	SetOptions = function(self, table)
		for name,value in pairs(table) do
			self:SetOption(name, value)
		end
	end;

	GetModule = function(self, name)
		local addonName, moduleName = splitstr(name, "/")
		local addon = self.addons[addonName]
		assert( addon ~= nil )
		local module = addon.modules[moduleName]
		assert( module ~= nil )
		return module;
	end;

	CreateAddon = function(self, name)
		if ( self.addons[name] ) then
			return self.addons[name];
		end

		local addon = {
			name = name;
			lib = self;
			version = GetAddOnMetadata(name, "Version");

			debug = false;

			modules = {};
		}
		setmetatable(addon, AddonClass)
		self.addons[name] = addon

		self.eventFrame:HookScript("OnEvent", function(self, event, ...)
			if (event == "ADDON_LOADED" and select(1, ...) == addon.name ) then
				-- initialize options
				if ( addon.lib.defaultOptions and not addon.lib.initialized ) then
					local options = addon.lib.savedVariable:Get() or {}

					-- SavedVariables have been loaded, create any new values
					for key,value in pairs(addon.lib.defaultOptions) do
						if ( options[key] == nil ) then
							options[key] = value;
						end
					end

					-- Remove any deprecated settings
					for key,value in pairs(options) do
						if ( addon.lib.defaultOptions[key] == nil )then
							options[key] = nil;
						end
					end

					addon.lib.savedVariable:Set(options)

					-- Fire OptionChanged event for all options
					for name,value in pairs(options) do
						addon.lib.callbacks.options:Fire(name, value)
					end

					addon.lib.initialized = true
				end

				-- initialize modules
				for _, module in pairs(addon.modules) do
					if ( not module.initialized ) then
						module.initialize()
						addon.lib.initialized = true
					end
				end
			end
		end)

		return addon
	end;
}
setmetatable(LibraryClass, DebugClass)
LibraryClass.__index = LibraryClass


---------------------------------------------
-- Initialize
---------------------------------------------
MissionMinder = {
	name = libname;
	ptr = GetBuildInfo():match('7[.]1[.][0-9]+');
	locale = LibStub("AceLocale-3.0"):GetLocale(libname);
	savedVariable = newGlobal(string.format("%sOptions", libname));
	eventFrame = CreateFrame("Frame");

	debug = false;
	initialized = false;

	addons = {};
	callbacks = {
		api = newCallbackRegistry();
		events = newCallbackRegistry();
		options = newCallbackRegistry();
	};
}
setmetatable(MissionMinder, LibraryClass)


MissionMinder.eventFrame:SetScript("OnEvent", function(self, event, ...)
	if (event ~= "ADDON_LOADED") then
		MissionMinder:DEBUG("[API]", event, ...)
		MissionMinder.callbacks.api:Fire(event, ...)
	end
end)
MissionMinder.eventFrame:RegisterEvent("ADDON_LOADED")
