local addon = MissionMinder:CreateAddon("MissionMinder")
local module = addon:CreateModule("missionPlanner")

---------------------------------------------
-- LOCALE
---------------------------------------------
local L = MissionMinder.locale
local L_1000 = L["K"]
local L_COMPLETED = L["completed"]
local L_WORKING = L["working"]
local L_IN_PARTY = L["in party"]
local L_AVAILABLE = L["available"]
local L_PARTY = L["Party"]
local L_COUNTERS = L["Counters"]
local L_NO_COUNTERS_AVAILABLE = L["No counters available"]
local L_ASSIGNED_MISSION = L["Assigned mission"]
local L_MINIMIZE_MISSION = L["Minimize mission"];

---------------------------------------------
-- DATA
---------------------------------------------
local minimizedParties = {} -- [missionID][1..3] = followerID

local followerMissions = {
	[LE_GARRISON_TYPE_6_0] = {}; -- [followerID] = mission
	[LE_GARRISON_TYPE_7_0] = {};
}


---------------------------------------------
-- Utility
---------------------------------------------
local RED_BIAS_PERCENTAGE = 0.10
local ORANGE_BIAS_PERCENTAGE = 0.50;
local WHITE_BIAS_PERCENTAGE = 1.00;


local IMPOSSIBLE_BIAS_COLOR = { 1, 0.1, 0.1 }
local HARD_BIAS_COLOR = { 1, 0.5, 0.25 }
local MENTORED_BIAS_COLOR = { 0.1, 1, 0.1 }
local NORMAL_BIAS_COLOR = { 1, 1, 1 }

local LEVELUP_ARROW_QUALITIES = {
	[LE_ITEM_QUALITY_COMMON] = "loottoast-arrow-green",
    [LE_ITEM_QUALITY_UNCOMMON]	= "loottoast-arrow-green",
	[LE_ITEM_QUALITY_RARE] = "loottoast-arrow-blue",
	[LE_ITEM_QUALITY_EPIC] = "loottoast-arrow-purple",
	[LE_ITEM_QUALITY_LEGENDARY] = "loottoast-arrow-orange"
}

local TRAIT_EXTRA_TRAINING = 80 -- whole party
local TRAIT_FAST_LEANER = 29 -- single follower
local TRAIT_HEARTHSTONE_PRO = 236 -- whole party
local expGainTraits = {
	[TRAIT_EXTRA_TRAINING] = 0.30,
	[TRAIT_FAST_LEANER] = 0.50,
	[TRAIT_HEARTHSTONE_PRO] = 0.35
}

local MINUTES = 60;
local HOURS = 60 * MINUTES;
local DAYS = 24 * HOURS;
local function getCompactDurationString(time)
	local str;
	if (time >= DAYS) then
		time = ((time - 1) / DAYS) + 1;
		return string.format(COOLDOWN_DURATION_DAYS, time);
	elseif (time >= HOURS) then
		time = ((time - 1) / HOURS) + 1;
		return string.format(COOLDOWN_DURATION_HOURS, time);
	elseif (time >= MINUTES) then
		time = ((time - 1) / MINUTES) + 1;
		return string.format(COOLDOWN_DURATION_MIN, time);
	else
		return string.format(COOLDOWN_DURATION_SEC, time);
	end
end

local function isEmpty(list)
	if (list) then
		for _,v in pairs(list) do
			if ( v ) then
				return false
			end
		end
	end
	return true
end


function MissionMinder_FollowerCanBeSentOnMission(follower)
	return follower and follower.isCollected and follower.status ~= GARRISON_FOLLOWER_INACTIVE;
end


local function updateRewardQualityBorder(frame, reward)
	if ( not frame.IconBorder ) then
		local iconborder = frame:CreateTexture("IconBorder", "OVERLAY")
		iconborder:SetTexture("Interface\\Common\\WhiteIconFrame")
		iconborder:SetAllPoints(frame.Icon)
		frame.IconBorder = iconborder
	end

    frame.IconBorder:Hide();
	local itemID = reward.itemID
	if ( itemID ) then
		local _, _, quality = GetItemInfo(itemID);
		if  ( quality and quality > LE_ITEM_QUALITY_COMMON and BAG_ITEM_QUALITY_COLORS[quality]) then
			frame.IconBorder:Show();
			frame.IconBorder:SetVertexColor(BAG_ITEM_QUALITY_COLORS[quality].r, BAG_ITEM_QUALITY_COLORS[quality].g, BAG_ITEM_QUALITY_COLORS[quality].b);
		end
	end
end

local function MissionButton_updateRewardQualityBorder(self, rewards, numRewards)
	if (numRewards > 0 and rewards) then
		local index = 1;
		for id, reward in pairs(rewards) do
			local frame = self.Rewards[index];
			if ( frame ) then -- Should never be nil, but if it ever is we ignore it
				updateRewardQualityBorder(frame, reward)
			end
			index = index + 1;
		end
	end
end

local function getFollowerMissionBiasRGB(missionID, followerID)
	local follower = C_Garrison.GetFollowerInfo(followerID);

    if ( not follower or follower.isTroop ) then
        return
    end

	local mentorLevel, mentorILevel = C_Garrison.GetPartyMentorLevels(missionID);
	local isMentored = (( mentorLevel and mentorLevel > follower.level )
			or (mentorILevel and mentorILevel > follower.iLevel ))

	local bias = C_Garrison.GetFollowerBiasForMission(missionID, followerID);
	if ( bias == -1 ) then
		return unpack(IMPOSSIBLE_BIAS_COLOR)
	elseif ( bias < 0 ) then
		return unpack(HARD_BIAS_COLOR)
	elseif ( isMentored ) then
		return unpack(MENTORED_BIAS_COLOR)
	else
		return unpack(NORMAL_BIAS_COLOR)
	end
end

local function getPotentialMissionXPGains(mission, follower)
	if ( not mission or not follower ) then
		return 0
	end

	local _, base = C_Garrison.GetMissionInfo(mission.missionID);

	local bonus = 0;
	for _, reward in pairs(C_Garrison.GetMissionRewardInfo(mission.missionID)) do
		if(reward.followerXP) then
			bonus = bonus + reward.followerXP
		end
	end

	local _, _, _, _, partyBuffs, _, modify = C_Garrison.GetPartyMissionInfo(mission.missionID);
	local gain = 0;
	for i = 1, #partyBuffs do
		if ( partyBuffs[i] ~= TRAIT_FAST_LEANER ) then
			gain = gain + (expGainTraits[partyBuffs[i] ] or 0);
		end
	end

	if ( follower.abilities ) then
		for i = 1, #follower.abilities do
			local ability = follower.abilities[i]
			if ( ability.isTrait and ability.id == TRAIT_FAST_LEANER ) then
				gain = gain + (expGainTraits[ability.id] or 0);
			end
		end
	end

	local bias = C_Garrison.GetFollowerBiasForMission(mission.missionID, follower.followerID);
	if ( mission.iLevel and mission.iLevel > 600 ) then
		if ( bias <= -1 ) then
			bias = ORANGE_BIAS_PERCENTAGE
		else
			bias = WHITE_BIAS_PERCENTAGE
		end
	else
		if ( bias <= -1 ) then
			bias = RED_BIAS_PERCENTAGE
		elseif ( bias < 0) then
			bias = ORANGE_BIAS_PERCENTAGE
		else
			bias = WHITE_BIAS_PERCENTAGE
		end
	end

	return bias * (base + modify + (bonus * (1 + gain)));
end


local function updateLevelUp(button, missionFrame)
	local missionPage = missionFrame:GetMissionPage()
	local mission = missionPage.missionInfo
	local follower = button.info

    if ( button.levelUp ) then
        button.levelUp:Hide();
    	if ( mission and missionPage:IsShown() and not missionFrame.FollowerTab:IsShown() and MissionMinder_FollowerCanBeSentOnMission(follower) and not follower.isTroop ) then
    		local maxXP = missionFrame.MissionComplete:GetFollowerNextLevelXP(follower.level, follower.quality);
    		local missionXPGains = getPotentialMissionXPGains(mission, follower);
    		if ( maxXP and follower.xp + missionXPGains >= maxXP ) then
    			button.levelUp:SetAtlas(LEVELUP_ARROW_QUALITIES[follower.quality], true);
    			button.levelUp:Show();
    		end
    	end
    end
end





---------------------------------------------
-- LandingPage
---------------------------------------------
function MissionMinder_LandingPageReport_AvailableMissions_UpdateRewardQualityBorder()
	local missions = GarrisonLandingPageReport.List.AvailableItems or {};
	local buttons = GarrisonLandingPageReport.List.listScroll.buttons;

	for i = 1, #buttons do
		local button = buttons[i]
		local mission = missions[button.id]
		if ( mission ) then
			local rewards = mission.rewards
			MissionButton_updateRewardQualityBorder(button, rewards, #rewards)
		end
	end
end

---------------------------------------------
-- FollowerList
---------------------------------------------
function MissionMinder_FollowerList_Initialize(self, func)
    if ( self.listScroll.buttons and #self.listScroll.buttons > 0 ) then
         func()
    else -- FollowerList has not been initialized yet, hook into :Setup
        hooksecurefunc(self, "Setup", func)
    end
end

function MissionMinder_FollowerList_UpdateFollowerLevelUp(self)
    local missionFrame = self:GetParent()
	local buttons = self.listScroll.buttons
	for i = 1, #buttons do
		updateLevelUp(buttons[i].Follower, missionFrame)
	end
end

function MissionMinder_FollowerList_UpdateXPRemaining(self)
    local missionFrame = self:GetParent()
	local buttons = self.listScroll.buttons

	for i = 1, #buttons do
        local button = buttons[i].Follower
    	local follower = button.info

        if ( button.xpRemaining ) then
            button.xpRemaining:SetText(nil)
        	if ( follower and follower.isCollected and not follower.isTroop ) then
        		local maxXP = missionFrame.MissionComplete:GetFollowerNextLevelXP(follower.level, follower.quality)
        		if ( maxXP ) then
        			button.xpRemaining:SetFormattedText(GARRISON_REWARD_XP_FORMAT, BreakUpLargeNumbers(maxXP - follower.xp))
        		end
        	end
        end
	end
end

function MissionMinder_FollowerList_UpdateBiasColorFollowerLevels(self)
    local missionFrame = self:GetParent()
    local missionPage = missionFrame:GetMissionPage()
	local mission = missionPage.missionInfo
    local buttons = self.listScroll.buttons

	for i = 1, #buttons do
		local button = buttons[i].Follower;
		local r, g, b = 1, 1, 1;

		if ( missionPage:IsShown() and not missionFrame.FollowerTab:IsShown() and mission and button.isCollected) then
			r, g, b = getFollowerMissionBiasRGB(mission.missionID, button.id);
		end

		button.PortraitFrame.Level:SetTextColor(r, g, b)
	end
end


local singleclickTimer = nil;
-- Override the standard OnClick event to delay just long enough to determine if it is a DoubleClick.
--If it is a double click then the OnClick event will not fire. If it isn't a double click then the
--OnClick event (which will run via timer) will still run; albeit slightly delayed.
function MissionMinder_FollowerList_SingleClick(self, button)
	if(singleclickTimer) then
		singleclickTimer:Cancel();
	end

	local missionFrame = self:GetParent():GetParent():GetParent():GetParent():GetParent()
	local missionPage = missionFrame:GetMissionPage()
	local followerTab = missionFrame.FollowerTab

	-- Only delay OnClick in very specific cases
	local mission = missionPage.missionInfo;
	if ( mission and missionPage:IsShown() and not followerTab:IsShown() ) then
		if ( self.isCollected and self.info.status == nil and button == "LeftButton" ) then
			singleclickTimer = C_Timer.NewTimer(0.2, function()
				self.OriginalOnClick(self, button)
			end);
			return;
		end
	end

	self.OriginalOnClick(self, button)
end

-- Handle OnDoubleClick event and disable the single click timer
function MissionMinder_FollowerList_AssignFollowerToMission(self, button)
	if(singleclickTimer) then
		singleclickTimer:Cancel();
	end

	local missionFrame = self:GetParent():GetParent():GetParent():GetParent():GetParent()
	local missionPage = missionFrame:GetMissionPage()
	local followerTab = missionFrame.FollowerTab

	-- Hide the rightclick menu
	GarrisonFollowerOptionDropDown.followerID = nil;
	CloseDropDownMenus();

	local mission = missionPage.missionInfo;
	if ( mission and missionPage:IsShown() and not followerTab:IsShown() ) then
		if ( self.isCollected and self.info.status == nil and button == "LeftButton" ) then
			module:DEBUGF("assignFollowerToMission(mission=%s, follower=%s)", mission.missionID, self.id);
			missionPage:AddFollower(self.id);
		end
	end
end

function MissionMinder_FollowerList_DisableSingleClickTimer()
	if(singleclickTimer) then
		singleclickTimer:Cancel()
		singleclickTimer = nil
	end
end

---------------------------------------------
-- MissionFrame
---------------------------------------------
function MissionMinder_MissionFrame_ClearAllParties(self)
	minimizedParties = {}
	local missions = C_Garrison.GetAvailableMissions(self.followerTypeID)
	for i = 1, #missions do
		local mission = missions[i];
		for i2 = 1, #mission.followers do
			local followerID = mission.followers[i2];
			module:DEBUGF("Removing follower %s from mission %d", followerID, mission.missionID)
			C_Garrison.RemoveFollowerFromMission(mission.missionID, followerID);
		end
	end
end

function MissionMinder_MissionFrame_UpdatePartyLevelUp(self)
	local partyFollowers = self:GetMissionPage().Followers;
	for i = 1, #partyFollowers do
		updateLevelUp(partyFollowers[i], self);
	end
end


---------------------------------------------
-- MissionFrame AvailalableMissionList
---------------------------------------------
function MissionMinder_AvailableMissionList_ShowTimeRemaining(self)
	--local self = OrderHallMissionFrame.MissionTab.MissionList;
	for _,button in pairs(self.listScroll.buttons) do
		local mission = button.info

		button.LocBG:SetVertexColor(1, 1, 1, 1);

		if ( mission and mission.offerEndTime and not self.showInProgress ) then
			local timeRemaining = mission.offerEndTime - GetTime();
			if ( timeRemaining > 0 ) then
				button.Expires:SetText(getCompactDurationString(timeRemaining))
				if ( timeRemaining < 2*HOURS ) then
					button.Expires:SetTextColor(1, 0, 0);
				else
					button.Expires:SetTextColor(1, 1, 1);
				end
				button.Expires:Show();
			else
				button.Expires:Hide();
			end
		else
			button.Expires:Hide()
		end
	end
end

function MissionMinder_AvailableMissionList_ShowMinimizedParty(self)
	if ( self.showInProgress ) then
		return
	end

	local missionFrame = self:GetParent():GetParent()
	for _,button in pairs(self.listScroll.buttons) do
		local mission = button.info

		if ( mission ) then
			local _, _, _, successChance = C_Garrison.GetPartyMissionInfo(mission.missionID);
			local party = minimizedParties[mission.missionID]

			if ( not isEmpty(party) and successChance ) then
				MissionMinder_UpdatePartyMembers(button, party, mission.numFollowers)
				button.SuccessChance:SetFormattedText(PERCENTAGE_STRING .. " " .. GARRISON_MISSION_CHANCE, successChance);
				button.SuccessChance:Show();
			else
				MissionMinder_UpdatePartyMembers(button, nil)
				button.SuccessChance:SetText(nil);
				button.SuccessChance:Hide();
			end
		end
	end
end

---------------------------------------------
-- MissionPage FollowerFrame
---------------------------------------------
local function followerWillMaxOnLevelUp(level, quality, maxLevel, maxQuality)
	--local maxLevel = missionFrame.followerMaxLevel;
	--local maxQuality = missionFrame.followerMaxQuality;

	local alreadyMaxLevel = level >= maxLevel
	local alreadyMaxQuality = quality >= maxQuality

	return (alreadyMaxQuality and level + 1 >= maxLevel) or (alreadyMaxLevel and quality + 1 >= maxQuality);
end

local function getActualMissionXPGains(missionFrame, potentialXP, follower)
	if( potentialXP > 0 ) then
		-- Maxed followers don't get any XP
		if ( not missionFrame.MissionComplete:GetFollowerNextLevelXP(follower.level, follower.quality) ) then
			return 0;
		end

		-- Follower expected to level and max, only show limited amount of XP they would gain
		local maxXP = missionFrame.MissionComplete:GetFollowerNextLevelXP(follower.level, follower.quality)
		if ( maxXP ) then
			local remainingXP = maxXP - follower.xp;
			if ( potentialXP >= remainingXP and followerWillMaxOnLevelUp(follower.level, follower.quality, missionFrame.followerMaxLevel, missionFrame.followerMaxQuality) ) then
				return remainingXP;
			end
		end
	end

	return potentialXP;
end

function MissionMinder_MissionPage_SpaceBarStartMission(self, key)
	self:SetPropagateKeyboardInput(true) -- Assume we should propagate keyboard input

	if ( key == "SPACE" ) then
		-- Don't propagate SPACE, we're going to consume it
		self:SetPropagateKeyboardInput(false)

		self.StartMissionButton:Click()
	end
end

function MissionMinder_MissionPage_UpdatePartyXPGains(self)
    local missionPage = self:GetMissionPage()
	local mission = missionPage.missionInfo;

	local partyFollowers = missionPage.Followers;
	for i = 1, #partyFollowers do
    	local follower = partyFollowers[i].info
		partyFollowers[i].XPGain:Hide();

		if ( follower and follower.isCollected and not follower.isTroop and mission ) then
	    	local potentialXP = getPotentialMissionXPGains(mission, follower)
	    	local missionXPGains = getActualMissionXPGains(self, potentialXP, follower);
	    	if (missionXPGains > 0) then
	    		partyFollowers[i].XPGain:SetFormattedText(XP_GAIN, BreakUpLargeNumbers(math.ceil(missionXPGains)));
	    		partyFollowers[i].XPGain:Show();
	    	end
		end
	end
end


function MissionMinder_MissionPageFollowerFrame_ClearFollower(self)
	local missionFrame = self:GetParent():GetParent():GetParent()
	local missionPage = missionFrame:GetMissionPage()

	local followerID = self.info and self.info.followerID or nil;
	self.info = nil;
	self.Name:Hide();
	if (self.Class) then
		self.Class:Hide();
	end
	self.PortraitFrame.Empty:Show();
	self.PortraitFrame.LevelBorder:SetAtlas("GarrMission_PortraitRing_LevelBorder");
	self.PortraitFrame.LevelBorder:SetWidth(58);
	self.PortraitFrame.Level:SetText("");
	for i = 1, #self.Counters do
		self.Counters[i]:Hide();
	end

	if (followerID) then
		if ( missionPage.missionInfo.numFollowers == 1 ) then
			missionPage.FollowerModel:ClearModel();
			missionPage.FollowerModel:Hide();
			missionPage.EmptyFollowerModel:Show();
		end
	end

--	GarrisonMissionPage_SetCounters();
end

function MissionMinder_MissionPageFollowerFrame_OnMouseUp(self, button)
	local missionFrame = self:GetParent():GetParent():GetParent()
	if ( button == "RightButton" ) then
		if ( self.info ) then
			missionFrame:RemoveFollowerFromMission(self, true)
		else
			missionFrame:GetMissionPage().MinimizeButton:Click()
		end
	end
end

---------------------------------------------
-- Minimize mission mixin
---------------------------------------------
MissionMinderMinimizeMixIn = {}

function MissionMinderMinimizeMixIn:ShowTooltip(...)
	GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
	GameTooltip:SetText(L_MINIMIZE_MISSION)
	GameTooltip:Show()
end

function MissionMinderMinimizeMixIn:MinimizeMission(...)
	local missionFrame = self:GetParent():GetParent():GetParent()
	local missionPage = missionFrame:GetMissionPage()
	local mission = missionPage.missionInfo
	if( not mission ) then
		return
	end

	if ( not minimizedParties[mission.missionID] ) then
		minimizedParties[mission.missionID] = {}
	end
	local party = minimizedParties[mission.missionID]

	for i = 1, #missionPage.Followers do
		local followerFrame = missionPage.Followers[i];
		local followerID = followerFrame.info and followerFrame.info.followerID or nil;
		MissionMinder_MissionPageFollowerFrame_ClearFollower(followerFrame)
		party[i] = followerID;
		if ( followerID ) then
			followerMissions[missionFrame.garrTypeID][followerID] = mission
		end

		module:DEBUGF("Saving follower [%d][%d] = %s", mission.missionID, i, tostring(party[i]));
	end

	--MissionMinder_MissionFrame_UpdatePartyLevelUp(missionFrame);
	--updatePartyXPGains();
	--missionPage.FollowerModel:Hide();
	--GarrisonMissionPage_UpdateEmptyString();

	missionPage:Hide();
	missionPage.missionInfo = nil;
	missionFrame.followerCounters = nil;
	missionFrame.followerTraits = nil;
	missionFrame.MissionTab.MissionList:Show();
end



---------------------------------------------
-- MissionPage load saved party
---------------------------------------------
function MissionMinder_LoadSavedParty(self)
	local missionFrame = self:GetParent():GetParent():GetParent():GetParent():GetParent()
	local missionPage = missionFrame:GetMissionPage()
	local mission = self.info

	if ( mission and minimizedParties[mission.missionID] ) then
		local followers = missionPage.Followers;
		local party = minimizedParties[mission.missionID];
		for i=1,#followers do
			local followerID = party[i];
			module:DEBUGF("Restoring follower [%d] = %s", i, tostring(followerID));

			if ( followerID ) then
				local followerInfo
				for k, v in pairs(missionFrame.FollowerList.followers) do
					if ( v.followerID == followerID ) then
						followerInfo = v --C_Garrison.GetFollowerInfo(v.garrFollowerID)
						break
					end
				end
				module:DEBUG(followerInfo)
				followers[i].info = followerInfo
				followers[i].Name:SetText(followerInfo.name);
				followers[i].Name:Show();
				if (followers[i].Class) then
					followers[i].Class:Show();
					followers[i].Class:SetAtlas(followerInfo.classAtlas);
				end
				followers[i].PortraitFrame.Empty:Hide();

				missionPage:UpdateFollowerModel(followerInfo);

				party[i] = nil
				followerMissions[missionFrame.garrTypeID][followerID] = nil
			end
			--updatePartyLevelUp()
			--updatePartyXPGains()
		end

		missionFrame:UpdateMissionParty(followers)
		missionFrame:UpdateMissionData(missionPage);
		missionFrame:GetFollowerBuffsForMission(missionPage.missionInfo.missionID);
		missionPage:SetCounters(missionPage.Followers, missionPage.Enemies, missionPage.missionInfo.missionID);
	end
end


function module.register(missionFrame)
	local missionPage = missionFrame:GetMissionPage()

	-- Create MissionPage.MinimizeButton
	CreateFrame("BUTTON", nil, missionPage, "MissionMinder_MissionPageMinimizeButton")


	--[[module:RegisterEvent("GARRISON_FOLLOWER_LIST_UPDATE", dirtyMissionCounterCache);
	module:RegisterEvent("GARRISON_FOLLOWER_XP_CHANGED", dirtyMissionCounterCache);
	module:RegisterEvent("GARRISON_FOLLOWER_REMOVED", dirtyMissionCounterCache);

	-- Mission tooltips
	local missionList = GarrisonMissionFrame.MissionTab.MissionList
	local missionButtons = missionList.listScroll.buttons;
	for i = 1, #missionButtons do
		local button = missionButtons[i]
		button:HookScript("OnEnter", createMissionTooltip);
	end
	hooksecurefunc("GarrisonMissionButton_OnEnter", createMissionTooltip)

	-- Periodically update mission tooltips
	for i = 1, #missionButtons do
		missionButtons[i].UpdateTooltip = GarrisonMissionButton_OnEnter
	end

	-- Update mission tooltip mechanics with minimized party counters
	GarrisonMission_DetermineCounterableThreats = determineCountersFromMinimizedParty
	]]
	-- Load minimized mission party
	-- MissionList is initialized early, no need to hook OrderHallMissionFrame:SetupMissionList()
	local missionButtons = missionFrame.MissionTab.MissionList.listScroll.buttons or {}
	for i = 1, #missionButtons do
		local button = missionButtons[i]
		button:HookScript("OnClick", MissionMinder_LoadSavedParty);
	end

	-- Show mission expiration
	for i = 1, #missionButtons do
		local button = missionButtons[i]
		button.Expires = button:CreateFontString("MissionExpires", "ARTWORK", "GameTooltipText")
		button.Expires:SetPoint("BOTTOM", button.MissionType);
		button.Expires:Show();
	end
	hooksecurefunc(missionFrame.MissionTab.MissionList, "Update", MissionMinder_AvailableMissionList_ShowTimeRemaining)

	-- Show minimized parties on mission button
	hooksecurefunc(missionFrame.MissionTab.MissionList, "Update", MissionMinder_AvailableMissionList_ShowMinimizedParty);

	-- Minimize mission when right click on MissionPage
	missionPage:SetScript("OnClick", function() missionPage.MinimizeButton:Click(); end);

	-- Minimize mission on FollowerFrame RightClick
	local partyFollowers = missionPage.Followers;
	for i = 1, #partyFollowers do
		partyFollowers[i]:SetScript("OnMouseUp", MissionMinder_MissionPageFollowerFrame_OnMouseUp)
	end

	-- Follower tooltips
	--[[local followerButtons = FollowerList.listScroll.buttons;
	for i = 1, #followerButtons do
		local button = followerButtons[i]
		button:SetScript("OnEnter", createFollowersTooltip);
		button:SetScript("OnLeave", hideMissionCountersTooltip);
	end]]

	-- Double click to assign followers to mission
	missionPage:HookScript("OnHide", MissionMinder_FollowerList_DisableSingleClickTimer)
	MissionMinder_FollowerList_Initialize(missionFrame.FollowerList, function()
		local followerButtons = missionFrame.FollowerList.listScroll.buttons
		for i = 1, #followerButtons do
			local button = followerButtons[i].Follower
			button.OriginalOnClick = button:GetScript("OnClick")
			button:SetScript("OnClick", MissionMinder_FollowerList_SingleClick);
			button:SetScript("OnDoubleClick", MissionMinder_FollowerList_AssignFollowerToMission);
		end
	end)

	-- Show remaining XP to level on follower list button
	hooksecurefunc(missionFrame.FollowerList, "UpdateData", MissionMinder_FollowerList_UpdateXPRemaining)
	MissionMinder_FollowerList_Initialize(missionFrame.FollowerList, function()
		local followerButtons = missionFrame.FollowerList.listScroll.buttons
		for i = 1, #followerButtons do
			local button = followerButtons[i].Follower
			local xpRemaining = button:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall");
			xpRemaining:SetPoint("BOTTOMRIGHT", button.BG);
			xpRemaining:SetText(nil);
			xpRemaining:Show();
			button.xpRemaining = xpRemaining;
		end
	end)

	-- Color follower levels by bias
	hooksecurefunc(missionFrame.FollowerList, "UpdateData", MissionMinder_FollowerList_UpdateBiasColorFollowerLevels)

	-- Change In Party follower busy frame color
	--[[hooksecurefunc("GarrisonFollowerList_Update", updateInPartyFollowerBusyFrameColor)

	-- Drag-to-swap party members
	module.GarrisonMissionPage_SetFollower = GarrisonMissionPage_SetFollower
	GarrisonMissionPage_SetFollower = setFollower
	GarrisonFollowerPlacerFrame:HookScript("OnClick", clearDragSource);
	for i = 1, #partyFollowers do
		partyFollowers[i]:HookScript("OnDragStart", setDragSource);
		partyFollowers[i]:HookScript("OnReceiveDrag", clearDragSource);
	end
	for i = 1, #followerButtons do
		local button = followerButtons[i]
		button:HookScript("OnDragStart", clearDragSource)
	end]]

	-- Follower level up
	hooksecurefunc(missionFrame.FollowerList, "UpdateData", MissionMinder_FollowerList_UpdateFollowerLevelUp)
	hooksecurefunc(missionFrame, "AssignFollowerToMission", MissionMinder_MissionFrame_UpdatePartyLevelUp);
	hooksecurefunc(missionFrame, "RemoveFollowerFromMission", MissionMinder_MissionFrame_UpdatePartyLevelUp);
	MissionMinder_FollowerList_Initialize(missionFrame.FollowerList, function()
		local followerButtons = missionFrame.FollowerList.listScroll.buttons
		for i = 1, #followerButtons do
			local button = followerButtons[i].Follower
			local levelUp = button.PortraitFrame:CreateTexture();
			levelUp:SetPoint("BOTTOM", button.PortraitFrame, "BOTTOMLEFT")
			levelUp:Hide();
			button.levelUp = levelUp
		end
	end)
	for i = 1, #partyFollowers do
		local levelUp = partyFollowers[i].PortraitFrame:CreateTexture();
		levelUp:SetPoint("BOTTOM", partyFollowers[i].PortraitFrame, "BOTTOMLEFT")
		levelUp:Hide();
		partyFollowers[i].levelUp = levelUp
	end

	-- Estimate party member XP gains
	hooksecurefunc(missionFrame, "AssignFollowerToMission", MissionMinder_MissionPage_UpdatePartyXPGains);
	hooksecurefunc(missionFrame, "RemoveFollowerFromMission", MissionMinder_MissionPage_UpdatePartyXPGains);
	for i = 1, #partyFollowers do
		local xpGain = partyFollowers[i]:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall")
		xpGain:SetPoint("TOP", partyFollowers[i], "BOTTOM")
		xpGain:SetTextColor(0, 1, 0)
		xpGain:SetText(nil)
		xpGain:Show()
		partyFollowers[i].XPGain = xpGain
	end

	-- Periodically update follower list (for mission time remaining)
	--[[FollowerList:SetScript("OnUpdate", followerList_OnUpdate);]]

	-- Show reward quality borders
	--	hooksecurefunc("GarrisonMissionButton_SetRewards", MissionButton_updateRewardQualityBorder);
	hooksecurefunc("GarrisonLandingPageReportList_UpdateItems", MissionMinder_LandingPageReport_AvailableMissions_UpdateRewardQualityBorder);
	GarrisonLandingPageReport.List.listScroll.update = GarrisonLandingPageReportList_UpdateItems
	--hooksecurefunc("GarrisonMissionPage_SetReward", updateRewardQualityBorder);

	-- Clear all parties when mission frame closes
	missionFrame:HookScript("OnHide", MissionMinder_MissionFrame_ClearAllParties)

	-- Change order of followers to better aid mission planning
	--[[	module.GarrisonFollowerList_SortFollowers = GarrisonFollowerList_SortFollowers
	GarrisonFollowerList_SortFollowers = sortFollowersForMission

	-- Update mission buttons for party
	hooksecurefunc("GarrisonMissionList_Update", updateMissionButtonsForParty)

	-- Update mission button rewards to include currencyQuantity
	hooksecurefunc("GarrisonMissionButton_SetRewards", setRewardCurrencyQuantity)
	]]

	missionPage:SetScript("OnKeyDown", MissionMinder_MissionPage_SpaceBarStartMission)

	--[[local debug = function(...) end
	module:RegisterEvent("GARRISON_LANDINGPAGE_SHIPMENTS", debug);
	module:RegisterEvent("GARRISON_SHOW_LANDING_PAGE", debug);
	module:RegisterEvent("GARRISON_HIDE_LANDING_PAGE", debug);
	module:RegisterEvent("GARRISON_BUILDING_ACTIVATABLE", debug);
	module:RegisterEvent("GARRISON_BUILDING_ACTIVATED", debug);
	module:RegisterEvent("GARRISON_ARCHITECT_OPENED", debug);
	module:RegisterEvent("GARRISON_MISSION_FINISHED", debug);
	module:RegisterEvent("GARRISON_MISSION_NPC_OPENED", debug);
	module:RegisterEvent("GARRISON_SHIPYARD_NPC_OPENED", debug);
	module:RegisterEvent("GARRISON_INVASION_AVAILABLE", debug);
	module:RegisterEvent("GARRISON_INVASION_UNAVAILABLE", debug);
	module:RegisterEvent("SHIPMENT_UPDATE", debug);
	module:RegisterEvent("ZONE_CHANGED_NEW_AREA", debug);
	module:RegisterEvent("ZONE_CHANGED", debug);
	module:RegisterEvent("ZONE_CHANGED_INDOORS", debug);
	module:RegisterEvent("GARRISON_FOLLOWER_CATEGORIES_UPDATED", debug)]]
end
